/* $Id$

This file is part of libmspgl
Copyright © 2007  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#include <msp/strings/formatter.h>
#include "except.h"
#include "material.h"
#include "mesh.h"
#include "object.h"
#include "objectinstance.h"
#include "objectpass.h"
#include "program.h"
#include "programdata.h"
#include "texture.h"
#include "texunit.h"

using namespace std;

namespace Msp {
namespace GL {

Object::Object():
	meshes(1, static_cast<Mesh *>(0)),
	main_texture(0),
	material(0)
{
	normal_pass=&passes[0];
}

Object::~Object()
{
	for(map<unsigned, ObjectPass>::iterator i=passes.begin(); i!=passes.end(); ++i)
		delete i->second.shdata;
}

bool Object::has_pass(const Tag &tag) const
{
	return passes.count(tag.id);
}

const ObjectPass &Object::get_pass(const Tag &tag) const
{
	map<unsigned, ObjectPass>::const_iterator i=passes.find(tag.id);
	if(i==passes.end())
		throw KeyError("Unknown pass");
	return i->second;
}

void Object::render(const Tag &tag) const
{
	render(get_pass(tag), 0);
}

void Object::render(const ObjectInstance &inst, const Tag &tag) const
{
	render(get_pass(tag), &inst);
}

void Object::setup_render(const ObjectPass &pass) const
{
	if(!meshes[0])
		throw InvalidState("Trying to render Object without mesh");

	if(pass.shprog)
	{
		pass.shprog->bind();
		pass.shdata->apply();
		for(unsigned i=0; i<textures.size(); ++i)
		{
			TexUnit::activate(i);
			textures[i]->bind();
		}
	}
	else if(main_texture && pass.use_textures)
		main_texture->bind();

	if(material)
		material->apply();
}

void Object::finish_render(const ObjectPass &pass) const
{
	if(pass.shprog)
	{
		Program::unbind();
		for(unsigned i=textures.size(); i--;)
		{
			TexUnit::activate(i);
			Texture::unbind();
		}
	}
	else if(main_texture)
		Texture::unbind();
}

void Object::render(const ObjectPass &pass, const ObjectInstance *inst) const
{
	setup_render(pass);

	if(inst)
		render_instance(pass, *inst);
	else
		meshes[0]->draw();

	finish_render(pass);
}

void Object::render_instance(const ObjectPass &pass, const ObjectInstance &inst) const
{
	inst.setup_render(pass);
	unsigned lod=min(inst.get_level_of_detail(), meshes.size()-1);
	meshes[lod]->draw();
	inst.finish_render(pass);
}


Object::Loader::Loader(Object &o, Collection &c):
	obj(o),
	coll(c)
{
	add("lod_mesh", &Loader::lod_mesh);
	add("material", &Object::material);
	add("material_inline", &Loader::material_inline);
	add("mesh",     &Loader::mesh);
	add("pass",     &Loader::pass);
	add("shader",   &Loader::shader);
	add("shader_texture", &Loader::shader_texture);
	add("texture",  &Loader::texture);
}

void Object::Loader::finish()
{
	for(map<unsigned, ObjectPass>::iterator i=obj.passes.begin(); i!=obj.passes.end(); ++i)
		if(i->second.shdata)
		{
			for(unsigned j=0; j<textures.size(); ++j)
				i->second.shdata->uniform(i->second.shprog->get_uniform_location(textures[j]), static_cast<int>(j));
		}
}

void Object::Loader::lod_mesh(unsigned l, const string &n)
{
	obj.meshes.resize(l+1, 0);
	obj.meshes[l]=coll.get<Mesh>(n);
}

void Object::Loader::material_inline()
{
	RefPtr<Material> mat=new Material;
	load_sub(*mat);
	coll.add(format("%p%p", &obj, mat.get()), mat.get());
	obj.material=mat.release();
}

void Object::Loader::mesh(const string &n)
{
	obj.meshes[0]=coll.get<Mesh>(n);
}

void Object::Loader::pass(const string &n)
{
	unsigned id=Tag(n).id;
	if(obj.passes.count(id))
		throw KeyError("Duplicate pass name");
	ObjectPass p;
	load_sub(p, coll);
	obj.passes[id]=p;
}

void Object::Loader::shader(const string &n)
{
	Program *shprog=coll.get<Program>(n);
	if(shprog)  // Allow for unsupported shaders
	{
		RefPtr<ProgramData> shdata=new ProgramData;
		load_sub(*shdata, *shprog);

		obj.normal_pass->shprog=shprog;
		if(obj.normal_pass->shdata)
			delete obj.normal_pass->shdata;
		obj.normal_pass->shdata=shdata.release();
	}
}

void Object::Loader::shader_texture(const string &n)
{
	unsigned eqsign=n.find('=');
	if(eqsign!=string::npos)
	{
		obj.textures.push_back(coll.get<Texture>(n.substr(eqsign+1)));
		textures.push_back(n.substr(0, eqsign));
	}
	else
	{
		obj.textures.push_back(coll.get<Texture>(n));
		textures.push_back(n);
	}
}

void Object::Loader::texture(const string &n)
{
	if(obj.main_texture)
		throw Exception("Only one main texture may be specified");

	obj.main_texture=coll.get<Texture>(n);
	obj.textures.push_back(obj.main_texture);
	textures.push_back("texture");
}

} // namespace GL
} // namespace Msp
