/* $Id$

This file is part of libmspgl
Copyright © 2007  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#ifndef MSP_GL_OBJECT_H_
#define MSP_GL_OBJECT_H_

#include <vector>
#include <msp/datafile/collection.h>
#include "objectpass.h"
#include "renderable.h"

namespace Msp {
namespace GL {

class Material;
class Mesh;
class ObjectInstance;
class Program;
class ProgramData;
class Texture;

/**
Stores data for a complete 3D object.  This includes a mesh, a shader and data
for it and a number of textures.  Only the mesh is mandatory, other components
are optional.

See also class ObjectInstance.
*/
class Object: public Renderable
{
private:
	std::vector<const Mesh *> meshes;
	std::vector<const Texture *> textures;
	const Texture *main_texture;
	std::map<unsigned, ObjectPass> passes;
	const Material *material;
	ObjectPass *normal_pass;

public:
	class Loader: public DataFile::Loader
	{
	public:
		typedef DataFile::Collection Collection;

	protected:
		Object &obj;
		Collection &coll;
	private:
		std::vector<std::string> textures;
	
	public:
		Loader(Object &, Collection &);

		Object &get_object() const { return obj; }
		Collection &get_collection() const { return coll; }
	private:
		virtual void finish();
		void lod_mesh(unsigned, const std::string &);
		void material_inline();
		void mesh(const std::string &);
		void pass(const std::string &);
		void shader(const std::string &);
		void shader_texture(const std::string &);
		void texture(const std::string &);
	};

	Object();
	~Object();

	virtual bool has_pass(const Tag &) const;
	const ObjectPass &get_pass(const Tag &) const;

	/**
	Renders the object.  A tag can be provided to render a non-default pass.
	*/
	virtual void render(const Tag &tag=Tag()) const;

	/**
	Renders the object with an instance.  The instance's hook functions are
	called before and after drawing the mesh.  A tag may also be given to render
	a non-default pass.
	*/
	virtual void render(const ObjectInstance &, const Tag &tag=Tag()) const;

	/**
	Renders multiple instances of the object in one go.  This may improve
	performance, as the object-specific render setup only has to be done once.
	Each instance's hook functions will be called before and after drawing the
	mesh.
	*/
	template<typename Iter>
	void render(Iter begin, Iter end, const Tag &tag=Tag()) const
	{ render(get_pass(tag), begin, end); }
private:
	void setup_render(const ObjectPass &) const;
	void finish_render(const ObjectPass &) const;
	void render(const ObjectPass &, const ObjectInstance *) const;

	template<typename Iter>
	void render(const ObjectPass &pass, Iter begin, Iter end) const
	{
		setup_render(pass);
		for(Iter i=begin; i!=end; ++i)
			render_instance(pass, **i);
		finish_render(pass);
	}

	void render_instance(const ObjectPass &, const ObjectInstance &) const;
};

} // namespace GL
} // namespace Msp

#endif
